from abc import ABC
from Shared.Exceptions import NeumSinkInfoEmptyException

class NeumSinkInfo(ABC):
    def __init__(self, number_vectors_stored:int) -> None:
        self.number_vectors_stored:str = number_vectors_stored

    def as_sink_info(dct:dict):
        if dct == None:
            raise NeumSinkInfoEmptyException("Received empty dict when converting to as_sink_info")
        return NeumSinkInfo(
            number_vectors_stored=dct.get("number_vectors_stored", None),
        )
    
    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['number_vectors_stored'] = self.number_vectors_stored
        return json_to_return


from enum import Enum

class LoaderEnum(str, Enum):
    autoloader = "autoloader"
    htmlloader = "htmlloader"
    markdownloader = "markdownloader"
    neumcsvloader = "neumcsvloader"
    neumjsonloader = "neumjsonloader"
    pdfloader = "pdfloader"

    def as_chunker_enum(loader_name: str):
        if loader_name == None or loader_name == "":
            return None
        try:
            enum_to_return = LoaderEnum[loader_name.lower()]
            return enum_to_return
        except KeyError as e:
            return None


from DataConnectors import (
    AzureBlobConnector,
    NeumFileConnector,
    NeumWebsiteConnector,
    S3Connector,
    SharepointConnector,
    SingleStoreConnector,
    SupabaseStorageConnector,
    PostgresConnector,
)

from DataConnectors.DataConnectorEnum import DataConnectorEnum

from Shared.Selector import Selector

from Shared.Exceptions import InvalidDataConnectorException

available_connectors = [enum.value for enum in list(DataConnectorEnum)]

def as_connector(dct:dict):
    if dct == None:
        raise InvalidDataConnectorException("Must supply a data connector configuration")
    if not isinstance(dct, dict):
        raise InvalidDataConnectorException("Data connector configuration needs to be a dictionary")
    
    connector_name = dct.get("connector_name", None)
    connector_name_enum = DataConnectorEnum.as_data_connector_enum(data_connector_name=connector_name)
    connector_information = dct.get("connector_information", None)
    selector = Selector.as_selector(dct.get("selector", None))

    if connector_name_enum == DataConnectorEnum.azureblobconnector:
        return AzureBlobConnector(connector_information=connector_information, selector=selector)
    elif connector_name_enum == DataConnectorEnum.neumwebsiteconnector:
        return NeumWebsiteConnector(connector_information=connector_information, selector=selector)
    elif connector_name_enum == DataConnectorEnum.neumfileconnector:
        return NeumFileConnector(connector_information=connector_information, selector=selector)
    elif connector_name_enum == DataConnectorEnum.postgresconnector:
        return PostgresConnector(connector_information=connector_information, selector=selector)
    elif connector_name_enum == DataConnectorEnum.s3connector:
        return S3Connector(connector_information=connector_information, selector=selector)
    elif connector_name_enum == DataConnectorEnum.sharepointconnector:
        return SharepointConnector(connector_information=connector_information, selector=selector)
    elif connector_name_enum == DataConnectorEnum.singlestoreconnector:
        return SingleStoreConnector(connector_information=connector_information, selector=selector)
    elif connector_name_enum == DataConnectorEnum.supabasestorageconnector:
        return SupabaseStorageConnector(connector_information=connector_information, selector=selector)
    else:
        raise InvalidDataConnectorException(f"{connector_name} is an invalid Data Connector. Available connectors: {available_connectors}] ")



from .TriggerSyncTypeEnum import TriggerSyncTypeEnum

class TriggerSchedule(object):
    def __init__(self, start_date: str, cadence: str, sync_type: TriggerSyncTypeEnum):
        self.start_date = start_date
        self.cadence = cadence
        self.sync_type = sync_type

    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API.

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['start_date'] = self.start_date
        json_to_return['cadence'] = self.cadence
        json_to_return['sync_type'] = self.sync_type
        return json_to_return
    
    def to_model(self):
        """Convert to model for serialization
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['start_date'] = self.start_date
        json_to_return['cadence'] = self.cadence
        json_to_return['sync_type'] = self.sync_type

        return json_to_return
    
    def as_trigger_schedule(dct:dict):
        if dct == None:
            return None
        return TriggerSchedule(
            start_date=dct.get("start_date", None),
            cadence=dct.get("cadence", None),
            sync_type=TriggerSyncTypeEnum.as_trigger_sync_type(dct.get("sync_type", None)),
        )


from psycopg2.extras import DictCursor
from datetime import datetime
from DataConnectors.DataConnector import DataConnector
from typing import List, Generator
from Shared.LocalFile import LocalFile
from Shared.CloudFile import CloudFile
from Shared.Exceptions import PostgresConnectionException
from decimal import Decimal
import psycopg2
import json

class PostgresConnector(DataConnector):
    """Postgres Connector \n
    connector_information requires:\n
    [ connection_string, query, batch_size]"""
    
    @property
    def connector_name(self) -> str:
        return "PostgresConnector"
    
    @property
    def required_properties(self) -> List[str]:
        return ["connection_string", "query"]

    @property
    def optional_properties(self) -> List[str]:
        return ["batch_size"]
    
    @property
    def available_metadata(self) -> str:
        return []

    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["NeumJSONLoader"]
    
    class CustomEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, Decimal):
                return str(obj)  # Convert Decimal to a string representation
            elif isinstance(obj, datetime):
                return obj.isoformat()  # Convert datetime to ISO format string
            return super().default(obj)
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        connection_string = self.connector_information['connection_string']
        query = self.connector_information['query']
        
        # Optional
        batch_size = self.connector_information.get('batch_size', 1000)

        with psycopg2.connect(connection_string) as connection:
            with connection.cursor(cursor_factory=DictCursor, name='pipeline_id') as cursor:
                cursor.itersize = 1000  # fetch 1000 rows at a time
                cursor.execute(query)
                batch_rows = []
                for row in cursor:
                    serialized_string = json.dumps(dict(row),cls=self.CustomEncoder)
                    serialized_dict = json.loads(serialized_string)
                    batch_rows.append(serialized_dict)
                    if(len(batch_rows) == batch_size):
                        yield CloudFile(data=json.dumps(batch_rows), metadata={}, id="Postgres")
                        batch_rows = []
                if len(batch_rows) > 0:
                    yield CloudFile(data=json.dumps(batch_rows), metadata={}, id="Postgres")

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # No metadatadata to determine what rows are new. Needs to be done through websocket
        yield from self.connect_and_list_full()

    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        data = json.loads(cloudFile.data)
        for row in data:
            yield LocalFile(in_mem_data=json.dumps(row), metadata=cloudFile.metadata)

    def validate(self) -> bool:
        try:
            connection_string = self.connector_information['connection_string']
            query = self.connector_information['query']
        except:
            raise ValueError(f"Required properties not set. Required properties: {self.required_properties}")
        
        if not all(x in self.available_metadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")
        
        try:
            psycopg2.connect(connection_string)
        except Exception as e:
            raise PostgresConnectionException(f"Connection to Postgres failed, check credentials. See Exception: {e}")      
        return True   


from .Loader import Loader
from .AutoLoader import AutoLoader
from .HTMLLoader import HTMLLoader
from .MarkdownLoader import MarkdownLoader
from .PDFLoader import PDFLoader
from .NeumCSVLoader import NeumCSVLoader
from .NeumJSONLoader import NeumJSONLoader
from .LoaderHelper import as_loader
from .LoaderEnum import LoaderEnum


from enum import Enum

class TriggerSyncTypeEnum(str, Enum):
    delta = "delta"
    full = "full"

    def as_trigger_sync_type(trigger_sync_type: str):
        if trigger_sync_type == None or trigger_sync_type == "":
            return None
        try:
            enum_to_return = TriggerSyncTypeEnum[trigger_sync_type.lower()]
            return enum_to_return
        except KeyError as e:
            return None


from typing import List
from SinkConnectors.SinkConnector import SinkConnector
from Shared.NeumSinkInfo import NeumSinkInfo
from Shared.NeumVector  import NeumVector
from Shared.NeumSearch import NeumSearchResult
from Shared.Exceptions import(
    SupabaseConnectionException,
    SupabaseInsertionException,
    SupabaseIndexInfoException,
    SupabaseQueryException
)
import vecs

class SupabaseSink(SinkConnector):
    """ Supabase Sink\n
    sink_information requires : [ database_connection ]"""

    @property
    def sink_name(self) -> str:
        return 'SupabaseSink'
    
    @property
    def required_properties(self) -> List[str]:
        return ['database_connection']

    @property
    def optional_properties(self) -> List[str]:
        return ['collection_name']

    def validation(self) -> bool:
        """Validate connector setup"""
        import vecs
        try:
            database_connection = self.sink_information['database_connection']
        except:
            raise ValueError(f"Required properties not set. Required properties: {self.required_properties}")
        try:
            vx = vecs.create_client(database_connection)
        except Exception as e:
            raise SupabaseConnectionException(f"Supabase connection couldn't be initialized. See exception: {e}")
        return True 

    def store(self, pipeline_id: str, vectors_to_store:List[NeumVector], task_id:str = "") -> int:
        database_connection = self.sink_information['database_connection']
        vx = vecs.create_client(database_connection)
        try:
            collection_name = self.sink_information.get("collection_name", f"pipeline_{pipeline_id}")
            dimensions = len(vectors_to_store[0].vector)
            db = vx.get_or_create_collection(name=collection_name, dimension=dimensions)
            to_upsert = []
            for i in range(0, len(vectors_to_store)):
                to_upsert.append((vectors_to_store[i].id, vectors_to_store[i].vector, vectors_to_store[i].metadata))

            db.upsert(records=to_upsert)
        except Exception as e:
            raise SupabaseInsertionException(f"Supabase storing failed. Exception {e}")
        finally:
            vx.disconnect()
        return len(vectors_to_store)
    
    def search(self, vector: List[float], number_of_results:int, pipeline_id:str) -> List:
        database_connection = self.sink_information['database_connection']
        vx = vecs.create_client(database_connection)
        collection_name = self.sink_information.get("collection_name", f"pipeline_{pipeline_id}")
        try:
            db = vx.get_collection(name=collection_name)
        except:
            raise SupabaseQueryException(f"Collection {collection_name} does not exist")
        finally:
            vx.disconnect()

        try:
            results = db.query(
                data=vector,
                include_metadata=True,
                include_value=True,
                limit=number_of_results,
            )
        except Exception as e:
            raise SupabaseQueryException(f"Error querying vectors from Supabase. Exception: {e}")
        finally:
            vx.disconnect()
        matches = []
        for result in results:
            matches.append(NeumSearchResult(
                id= str(result[0]),
                metadata=result[2],
                score=result[1]
            ))
        
        vx.disconnect()
        return matches
    
    def info(self, pipeline_id: str) -> NeumSinkInfo:
        database_connection = self.sink_information['database_connection']
        vx = vecs.create_client(database_connection)
        collection_name = self.sink_information.get("collection_name", f"pipeline_{pipeline_id}")
        try:
            db = vx.get_collection(name=collection_name)
        except:
            raise SupabaseIndexInfoException(f"Collection {collection_name} does not exist")
        finally:
            vx.disconnect()
        
        number_of_vectors = db.table.select('count(*)')[0].count

        vx.disconnect()
        return NeumSinkInfo(number_vectors_stored=number_of_vectors)


from Shared.NeumSinkInfo import NeumSinkInfo
from Shared.NeumVector  import NeumVector
from Shared.NeumSearch import NeumSearchResult
from Shared.Exceptions import(
    SinglestoreConnectionException,
    SinglestoreInsertionException,
    SinglestoreIndexInfoException,
    SinglestoreQueryException
)
from SinkConnectors.SinkConnector import SinkConnector
from typing import List

import singlestoredb as s2

class SingleStoreSink(SinkConnector):
    """ SingleStore Sink\n
    sink_information requires : [ 'url', 'api_key' ]"""
        
    @property
    def sink_name(self) -> str:
        return 'SingleStoreSink'
    
    @property
    def required_properties(self) -> List[str]:
        return ['url', 'api_key']

    @property
    def optional_properties(self) -> List[str]:
        return ['collection_name']

    def validation(self) -> bool:
        """Validate connector setup"""
        import singlestoredb as s2
        try:
            url = self.sink_information['url']
            table = self.sink_information['table']
        except:
            raise ValueError(f"Required properties not set. Required properties: {self.required_properties}")
        try: 
            s2.connect(url)
        except Exception as e:
            raise SinglestoreConnectionException(f"There was a problem connecting to Singlestore. See Exception: {e}")
        return True 

    def store(self, pipeline_id: str, vectors_to_store:List[NeumVector], task_id:str = "") -> int:
        batch_size = 1000
        url = self.sink_information['url']
        table = self.sink_information['table']

        # Get metadata list
        # metadata_fields = ""
        if vectors_to_store[0].metadata is not None:
            metadata_keys = vectors_to_store[0].metadata.keys()
            if(len(metadata_keys) > 0):
                metadata_fields = ', ' + ', '.join(metadata_keys)
            else: metadata_fields = ""

        # Determine the number of batches
        num_batches = (len(vectors_to_store) + batch_size-1) // batch_size
        try:
            with s2.connect(url) as conn:
                with conn.cursor() as cur:
                    for batch_num in range(num_batches):
                        # Get the start and end indices for this batch
                        start_idx = batch_num * 1000
                        end_idx = start_idx + 1000
                        # Generate the values string for this batch
                        # Assumes a string based id. Needs to ensure the table is created as such
                        if metadata_fields != "":
                            value_strings = [
                                "(\'{}\', \'{}\', json_array_pack(\'{}\'), {})".format(
                                    vectors_to_store[i].id, vectors_to_store[i].vector, ', '.join(["'" + str(value) + "'" for value in vectors_to_store[i].metadata.values()])
                                )
                                for i in range(start_idx, min(end_idx, len(vectors_to_store)))
                            ]
                        else:
                            value_strings = [
                                "(\'{}\', \'{}\', json_array_pack(\'{}\'))".format(
                                    vectors_to_store[i].id, vectors_to_store[i].vector
                                )
                                for i in range(start_idx, min(end_idx, len(vectors_to_store)))
                            ]
                        values_sql = ',\n'.join(value_strings)
                        insert_query = f"""INSERT INTO {table} (id, vector{metadata_fields})\nVALUES\n{values_sql};"""
                        cur.execute(insert_query)
        except Exception as e:
            raise SinglestoreInsertionException("SingleStore storing failed. Try again later.")
        
        return len(vectors_to_store), None
    
    def search(self, vector: List[float], number_of_results: int, pipeline_id: str) -> List[NeumSearchResult]:
        url = self.sink_information['url']
        table = self.sink_information['table']

        query = f"""SELECT id, text, dot_product(vector, json_array_pack('{vector}')) AS score
        FROM {table}
        ORDER BY score DESC
        LIMIT {number_of_results}"""
        
        try:
            with s2.connect(url, results_type="dict") as conn:
                with conn.cursor() as cur:
                    matches:List[NeumSearchResult] = []
                    cur.execute(query)
                    for row in cur.fetchall():
                        matches.append(NeumSearchResult(id = str(dict(row)['id']), metadata={"text": str(dict(row)['text'])}, score=dict(row)['score']))
                    return matches
        except Exception as e:
            raise SinglestoreQueryException(f"Failed to query single store. Exception - {e}")

    def info(self, pipeline_id: str) -> NeumSinkInfo:
        url = self.sink_information['url']
        table = self.sink_information['table']

        query = f"""SELECT Count(*) as count
        FROM {table}"""
        
        try:
            with s2.connect(url, results_type="dict") as conn:
                with conn.cursor() as cur:
                    cur.execute(query)
                    rows = cur.fetchall()
                    return NeumSinkInfo(number_vectors_stored=dict(rows[0])["count"])
        except Exception as e:
            raise SinglestoreIndexInfoException(f"Failed to get info for singlestore. Exception - {e}") 


from datetime import datetime
from typing import List, Generator
import boto3
from Shared.LocalFile import LocalFile
from Shared.CloudFile import CloudFile
from Shared.Exceptions import S3ConnectionException
from DataConnectors.DataConnector import DataConnector
import tempfile
import os

class S3Connector(DataConnector):
    """" Neum File Connector \n
    connector_information contains: \n
    [ aws_key_id, aws_access_key, bucket_name ]"""

    @property
    def connector_name(self) -> str:
        return "S3Connector"
    
    @property
    def required_properties(self) -> List[str]:
        return ["aws_key_id", "aws_access_key", "bucket_name"]

    @property
    def optional_properties(self) -> List[str]:
        return ["prefix"]
    
    @property
    def available_metadata(self) -> str:
        return ["key" , "last_modified", "metadata" ]
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["AutoLoader", "HTMLLoader", "MarkdownLoader", "NeumCSVLoader", "NeumJSONLoader", "PDFLoader"]
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        aws_key_id= self.connector_information['aws_key_id']
        aws_access_key= self.connector_information['aws_access_key']
        bucket_name= self.connector_information['bucket_name']
        prefix = self.connector_information.get("prefix", "")
        session = boto3.Session(
            aws_access_key_id=aws_key_id,
            aws_secret_access_key=aws_access_key,
        )
        s3_resource = session.resource('s3')
        s3_client = session.client("s3")
        bucket = s3_resource.Bucket(bucket_name)

        # List out the files to be passed on
        for obj in bucket.objects.filter(Prefix=prefix):
            # Hard coded metadata available on the object
            metadata = {
                "last_modified" : obj.last_modified
            }
            selected_metadata  = {k: metadata[k] for k in self.selector.to_metadata if k in metadata}
            # If metadata passed, then I will add all the user generated values that are associated to the file
            if "metadata" in self.selector.to_metadata:
                # Make an additional call to get the full context
                additional_metdata:dict = s3_client.head_object(Bucket=bucket_name, Key=obj.key)
                selected_metadata.update(additional_metdata['Metadata'])
            yield CloudFile(file_identifier=obj.key, metadata=selected_metadata, id=obj.key)

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # Connect to S3
        aws_key_id= self.connector_information['aws_key_id']
        aws_access_key= self.connector_information['aws_access_key']
        bucket_name= self.connector_information['bucket_name']
        prefix = self.connector_information.get("prefix", "")   
        session = boto3.Session(
            aws_access_key_id=aws_key_id,
            aws_secret_access_key=aws_access_key,
        )
        s3_resource = session.resource('s3')
        s3_client = session.client("s3")
        bucket = s3_resource.Bucket(bucket_name)

        # List out the files that have changed
        for obj in bucket.objects.filter(Prefix=prefix):    
            # Check if file has changed
            if(last_run < obj.last_modified):
                # If file changed, then download
                metadata = {
                    "key" : obj.key,
                    "last_modified" : obj.last_modified
                }
                selected_metadata  = {k: metadata[k] for k in self.selector.to_metadata if k in metadata}
                # If metadata passed, then I will add all the user generated values that are associated to the file
                if "metadata" in self.selector.to_metadata:
                    # Make an additional call to get the full context
                    additional_metdata:dict = s3_client.head_object(Bucket=bucket_name, Key=obj.key)
                    selected_metadata.update(additional_metdata['Metadata'])
                yield CloudFile(file_identifier=obj.key, metadata=selected_metadata, id=obj.key)

    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        # Connect to S3
        aws_key_id= self.connector_information['aws_key_id']
        aws_access_key= self.connector_information['aws_access_key']
        bucket_name= self.connector_information['bucket_name']
        session = boto3.Session(
            aws_access_key_id=aws_key_id,
            aws_secret_access_key=aws_access_key,
        )
        s3_client = session.client("s3")
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = f"{temp_dir}/{cloudFile.file_identifier}"
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            s3_client.download_file(bucket_name, cloudFile.file_identifier, file_path)
            yield LocalFile(file_path=file_path, metadata=cloudFile.metadata, id=cloudFile.id)

    def validate(self) -> bool:
        try:
            aws_key_id= self.connector_information['aws_key_id']
            aws_access_key= self.connector_information['aws_access_key']
            bucket_name= self.connector_information['bucket_name']
        except:
            raise ValueError(f"Required properties not set. Required properties: {self.required_properties}")
        
        if not all(x in self.available_metadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")
        
        try:
            session = boto3.Session(
                aws_access_key_id=aws_key_id,
                aws_secret_access_key=aws_access_key,
            )
            client = session.client("s3")
            client.head_bucket(Bucket=bucket_name)
        except Exception as e:
            raise S3ConnectionException(f"Connection to S3 failed, check key and key ID. See Exception: {e}")
        return True 




from .Chunker import Chunker
from .CharacterChunker import CharacterChunker
from .CustomChunker import CustomChunker
from .RecursiveChunker import RecursiveChunker
from .ChunkerHelper import as_chunker
from .ChunkerEnum import ChunkerEnum


from typing import List, Generator
from Shared.NeumDocument import NeumDocument
from Shared.LocalFile import LocalFile
from Loaders.Loader import Loader
import json

class NeumJSONLoader(Loader):
    """" Neum JSON Loader """

    @property
    def required_properties(self) -> List[str]:
        return []

    @property
    def optional_properties(self) -> List[str]:
        return ["id_key"]

    @property
    def loader_name(self) -> str:
        return "NeumJSONLoader"
    
    @property
    def available_metadata(self) -> List[str]:
        return ["custom"]

    @property
    def availableContent(self) -> List[str]:
        return ["custom"]
    
    def validate(self) -> bool:
        return True   

    def load(self, file: LocalFile) -> Generator[NeumDocument, None, None]:
        """Load data into Document objects."""
        id_key = self.loader_information.get('id_key', "id")

        json_data = None
        if file.file_path:
            with open(file.file_path, 'r') as json_file:
                json_data = json.load(json_file)
        elif file.in_mem_data:
            json_data = json.loads(file.in_mem_data)

        if json_data is not None:
            processed_json = self.process_item(item=json_data, id_key=id_key)

            for item in processed_json:
                content = ''.join(item['data'])
                metadata = item['metadata']
                document_id = item['id']
                metadata.update(file.metadata)
                yield NeumDocument(content=content, metadata=metadata, id=document_id)

    def process_item(self, item, prefix="", metadata=None, document_id=None, id_key="id"):
        if metadata is None:
            metadata = {}

        if isinstance(item, dict):
            new_metadata = self.extract_metadata(item, self.selector.to_metadata)
            new_metadata.update(metadata)
            result = []
            for key, value in item.items():
                new_prefix = f"{prefix}.{key}" if prefix else key
                new_document_id = f"{item.get(id_key, '')}.{new_prefix}" if id_key in item else new_prefix
                if self.selector.to_embed is None or new_prefix in self.selector.to_embed or not self.selector.to_embed:
                    result.extend(self.process_item(item=value, prefix=new_prefix, metadata=new_metadata, document_id=new_document_id, id_key=id_key))
            return result
        elif isinstance(item, list):
            result = []
            for value in item:
                result.extend(self.process_item(item=value, prefix=prefix, metadata=metadata, document_id=document_id, id_key=id_key))
            return result
        else:
            return [{'data': [f"{item}"], 'metadata': metadata, 'id': document_id or prefix}]

    def extract_metadata(self, item: dict, metadata_keys: List[str]) -> dict:
        return {key: item[key] for key in metadata_keys if key in item}


from .DataConnector import DataConnector
from .AzureBlobConnector import AzureBlobConnector
from .NeumWebsiteConnector import NeumWebsiteConnector
from .S3Connector import S3Connector
from .SupabaseStorageConnector import SupabaseStorageConnector
from .PostgresConnector import PostgresConnector
from .SharepointConnector import SharepointConnector
from .SingleStoreConnector import SingleStoreConnector
from .NeumFileConnector import NeumFileConnector
from .ConnectorHelper import as_connector
from .DataConnectorEnum import DataConnectorEnum


from DataConnectors.DataConnector import DataConnector
from typing import List, Generator
from Shared.LocalFile import LocalFile
from Shared.CloudFile import CloudFile
from Shared.Exceptions import WebsiteConnectionException
from bs4 import BeautifulSoup
import tempfile
import requests

class NeumWebsiteConnector(DataConnector):
    """" Neum Website Connector \n
    connector_information contains: [ url ]\n
    available metadata: [ url ]\n
    available content: [ website ]"""
  
    @property
    def connector_name(self) -> str:
        return "NeumWebsiteConnector"
    
    @property
    def required_properties(self) -> List[str]:
        return ["url"]

    @property
    def optional_properties(self) -> List[str]:
        return []
    
    @property
    def available_metadata(self) -> str:
        return ['url']
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["HTMLLoader"]
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        # Send an HTTP GET request to the website
        url:str = str(self.connector_information['url'])
        clean_url = url.replace(" ", "")
        list_urls = clean_url.split(" , ")
        for u in list_urls:
            available_metadata = {'url':u}
            selected_metadata  = {k: available_metadata[k] for k in self.selector.to_metadata if k in available_metadata}
            yield CloudFile(file_identifier=u, metadata=selected_metadata, id=u)

    def connect_and_list_delta(self) -> Generator[CloudFile, None, None]:
        yield from self.connect_and_list_full()
    
    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
            response = requests.get(cloudFile.file_identifier)
            # Parse the HTML content
            soup = BeautifulSoup(response.content, 'html.parser')
            # Find the <body> element and extract its HTML content
            body = soup.find('body')
            body_html = str(body)  # Convert the body tag to a string to get its HTML content
            # Create a temporary file and write the extracted HTML to it
            with tempfile.NamedTemporaryFile(delete=False, suffix=".html", mode='w', encoding="utf-8") as temp:
                temp.write(body_html)
                yield LocalFile(file_path=temp.name, metadata=cloudFile.metadata, id=cloudFile.id)
        
    def validate(self) -> bool:
        # Check for required properties
        try:
            url:str = str(self.connector_information['url'])
        except:
            raise ValueError(f"Required properties not set. Required properties: {self.required_properties}")
        
        # Check for metadata values
        if not all(x in self.available_metadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")
        
        # Check to see that site exists
        try:
            requests.get(url)
        except Exception as e:
            raise WebsiteConnectionException(f"Connection to website failed, check url. See Exception: {e}")      
        return True 


from .Pipeline import (
    Pipeline
)

from .PipelineRun import (
    PipelineRun,
    PipelineRunStatus, 
    PipelineRunTaskDetails
)
from .TriggerSchedule import TriggerSchedule
from .TriggerSyncTypeEnum import TriggerSyncTypeEnum


from typing import List, Generator
from Shared.NeumDocument import NeumDocument
from Shared.LocalFile import LocalFile
from Loaders.Loader import Loader
from langchain.document_loaders import UnstructuredMarkdownLoader

class MarkdownLoader(Loader):
    """" Markdown Loader """
    """" loader_information contains: """
    """ [ ] """

    @property
    def loader_name(self) -> str:
        return "MarkdownLoader"
    
    @property
    def required_properties(self) -> List[str]:
        return []

    @property
    def optional_properties(self) -> List[str]:
        return []
    
    @property
    def available_metadata(self) -> List[str]:
        return []

    @property
    def availableContent(self) -> List[str]:
        return []

    #Probably worth re-writing directly on top of pypdf to get access
    #to more metadata including images, tables, etc.
    def load(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        """Load data into Document objects."""
        loader = UnstructuredMarkdownLoader(file_path=file.file_path)
        documents = loader.load()
        for doc in documents:
            yield NeumDocument(id=file.id, content=doc.page_content, metadata=file.metadata)

    def validate(self) -> bool:
        return True   


from typing import List, Generator
from Chunkers.Chunker import Chunker
from Shared.NeumDocument import NeumDocument
from neumai_tools import semantic_chunking

class CustomChunker(Chunker):
    """" Custom Chunker \n
    chunker_information requires: \n
    [ code ] \n
    chunker_information optional: \n
    [ batch_size ]"""
    
    @property
    def chunker_name(self) -> str:
        return "CustomChunker"
    
    @property
    def required_properties(self) -> List[str]:
        return ["code"]

    @property
    def optional_properties(self) -> List[str]:
        return ["batch_size"]

    def chunk(self, documents:List[NeumDocument]) -> Generator[List[NeumDocument], None, None]:
        
        chunking_code_exec=self.chunker_information['code']
        batch_size = self.chunker_information.get('batch_size', 1000)
        
        # Probably add some code to check the code I am about to run.
        # Code format must follow:
        # def split_text_into_chunks(text) -> List[NeumDocument]:

        # Iterate through documents to chunk them and them merge them back up
        documents_to_embed:List[NeumDocument] = []
        for doc in documents:
            chunks = semantic_chunking(documents=[doc], chunking_code_exec=chunking_code_exec)
            for i in range(len(chunks)):
                documents_to_embed.append(NeumDocument(id=doc.id + "_" + str(i), content=chunks[i].page_content, metadata=doc.metadata))
                if(len(documents_to_embed) == batch_size):
                    yield documents_to_embed
                    documents_to_embed = []
        
        # Pass last remaining
        if(len(documents_to_embed) > 0):
            yield documents_to_embed

    def validate(self) -> bool:
        return True   


from .ReplicateEmbed import ReplicateEmbed
from .AzureOpenAIEmbed import AzureOpenAIEmbed
from .EmbedConnector import EmbedConnector
from .OpenAIEmbed import OpenAIEmbed
from .EmbedConnector import EmbedConnector
from .EmbedHelper import as_embed


from enum import Enum

class ChunkerEnum(str, Enum):
    characterchunker = "characterchunker"
    customchunker = "customchunker"
    recursivechunker = "recursivechunker"

    def as_chunker_enum(chunker_name: str):
        if chunker_name == None or chunker_name == "":
            return None
        try:
            enum_to_return = ChunkerEnum[chunker_name.lower()]
            return enum_to_return
        except KeyError as e:
            return None


from abc import ABC
from Shared.Exceptions import LocalFileEmptyException

class LocalFile(ABC):
    def __init__(self, metadata:dict, file_path:str = None, in_mem_data:dict = None, type:str = "Any", id:str = None) -> None:
        self.file_path:str = file_path
        self.metadata:dict = metadata
        self.in_mem_data:dict = in_mem_data
        self.type:str = type
        self.id:str = id

    def as_file(dct:dict):
        if dct == None:
            raise LocalFileEmptyException("Received empty dict when converting to as_file")
        return LocalFile(
            file_path=dct.get("file_path", None),
            metadata=dct.get("metadata", None),
            in_mem_data=dct.get("in_mem_data", None),
            type=dct.get("type", None),
            id=dct.get("id", None)
        )
    
    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['file_path'] = self.file_path
        json_to_return['metadata'] = self.metadata
        json_to_return['in_mem_data'] = self.in_mem_data
        json_to_return['type'] = self.type
        json_to_return['id'] = self.id
        return json_to_return
    


from enum import Enum

class EmbedConnectorEnum(str, Enum):
    openaiembed = "openaiembed"
    azureopenaiembed = "azureopenaiembed"
    replicateembed = "replicateembed"

    def as_embed_connector_enum(embed_connector_name: str):
        if embed_connector_name == None or embed_connector_name == "":
            return None
        try:
            enum_to_return = EmbedConnectorEnum[embed_connector_name.lower()]
            return enum_to_return
        except KeyError as e:
            return None


from datetime import datetime
from typing import List, Generator
from Shared.LocalFile import LocalFile
from Shared.CloudFile import CloudFile
from Shared.Exceptions import SharepointConnectionException
from DataConnectors.DataConnector import DataConnector
import tempfile
import requests

class SharepointConnector(DataConnector):
    """" Sharepoint Connector \n
    connector_information requires:\n
    [ tenant_id, client_id, client_secret, site_id ]"""
    
    @property
    def connector_name(self) -> str:
        return "SharepointConnector"
    
    @property
    def required_properties(self) -> List[str]:
        return ["tenant_id", "client_id", "client_secret", "site_id"]

    @property
    def optional_properties(self) -> List[str]:
        return []

    @property
    def available_metadata(self) -> str:
        return ["createdDateTime", "lastModifiedDateTime", "name", "createdBy.user.email", "createdBy.user.id", "createdBy.user.displayName", "lastModifiedBy.user.email", "lastModifiedBy.user.id", "lastModifiedBy.user.displayName"]
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["AutoLoader", "HTMLLoader", "MarkdownLoader", "NeumCSVLoader", "NeumJSONLoader", "PDFLoader"]
    
    def process_folder(self, site_id:str, drive_id:str, folder_id:str, headers:dict) -> Generator[CloudFile, None, None]:
        import requests
        folder_url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives/{drive_id}/items/{folder_id}/children"
        folder_response = requests.get(folder_url, headers=headers)
        folder_info = folder_response.json()
        folder_items =  folder_info.get('value', [])
        for item in folder_items:
            if 'file' in item.keys():
                # It's a file, download it
                file_url = item['@microsoft.graph.downloadUrl']
                file_name = item['name']
                file_type = item['file']['mimeType']
                available_metadata = {
                    "createdDateTime":item['createdDateTime'],
                    "lastModifiedDateTime":item['lastModifiedDateTime'],
                    "name":item['name'],
                    "createdBy.user.email":item['createdBy']['user']['email'],
                    "createdBy.user.id":item['createdBy']['user']['id'],
                    "createdBy.user.displayName":item['createdBy']['user']['displayName'],
                    "lastModifiedBy.user.email":item['createdBy']['user']['email'],
                    "lastModifiedBy.user.id":item['createdBy']['user']['id'],
                    "lastModifiedBy.user.displayName":item['createdBy']['user']['displayName'],
                }
                selected_metadata  = {k: available_metadata[k] for k in self.selector.to_metadata if k in available_metadata}
                yield CloudFile(file_identifier=file_url, id=file_name, type=file_type, metadata=selected_metadata)
            
            elif 'folder' in item.keys():
                # It's a folder, process it recursively
                yield from self.process_folder(site_id=site_id, drive_id=drive_id, folder_id=item['id'], headers=headers)
    
    def process_folder_delta(self, site_id:str, drive_id:str, folder_id:str, headers:dict, metadata_keys:List[str], last_run:datetime) -> Generator[CloudFile, None, None]:
        import requests
        folder_url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives/{drive_id}/items/{folder_id}/children"
        folder_response = requests.get(folder_url, headers=headers)
        folder_info = folder_response.json()
        folder_items =  folder_info.get('value', [])
        for item in folder_items:
            if 'file' in item.keys():
                # It's a file, download it
                file_url = item['@microsoft.graph.downloadUrl']
                file_name = item['name']
                file_type = item['file']['mimeType']
                available_metadata = {
                    "createdDateTime":item['createdDateTime'],
                    "lastModifiedDateTime":item['lastModifiedDateTime'],
                    "name":item['name'],
                    "createdBy.user.email":item['createdBy']['user']['email'],
                    "createdBy.user.id":item['createdBy']['user']['id'],
                    "createdBy.user.displayName":item['createdBy']['user']['displayName'],
                    "lastModifiedBy.user.email":item['createdBy']['user']['email'],
                    "lastModifiedBy.user.id":item['createdBy']['user']['id'],
                    "lastModifiedBy.user.displayName":item['createdBy']['user']['displayName'],
                }
                if last_run < item['lastModifiedDateTime']:
                    selected_metadata  = {k: available_metadata[k] for k in metadata_keys if k in available_metadata}
                    yield CloudFile(file_identifier=file_url, id=file_name, type=file_type, metadata=selected_metadata)
            
            elif 'folder' in item.keys():
                # It's a folder, process it recursively
                yield from self.process_folder(site_id=site_id, drive_id=drive_id, folder_id=item['id'], headers=headers)

    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        """Connect to source and download file into local storage"""
        import requests

        #Required
        tenant_id = self.connector_information['tenant_id']
        client_id = self.connector_information['client_id']
        client_secret = self.connector_information['client_secret']
        site_id = self.connector_information['site_id']

        #Auth
        token_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
        token_data = {
            'grant_type': 'client_credentials',
            'client_id': client_id,
            'client_secret': client_secret,
            'scope': 'https://graph.microsoft.com/.default'
        }
        token_r = requests.post(token_url, headers={"Content-Type":"application/x-www-form-urlencoded"}, data=token_data)
        token = token_r.json().get('access_token')

        # Get all document libraries in the site
        headers = {
            'Authorization': f'Bearer {token}',
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        doc_libraries_url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives"
        doc_libraries_response = requests.get(doc_libraries_url, headers=headers)
        doc_libraries_info = doc_libraries_response.json()

        # Iterate through all document libraries
        for doc_library in doc_libraries_info.get('value', []):
            doc_library_id = doc_library['id']
            yield from self.process_folder(site_id=site_id, drive_id=doc_library_id, folder_id='root', headers=headers)

    def connect_and_list_delta(self, last_run:datetime) -> Generator[LocalFile, None, None]:
        """Check for changes in the source"""
        """Code to be pushed to a worker and run on a schedule"""
        import requests

        #Required
        tenant_id = self.connector_information['tenant_id']
        client_id = self.connector_information['client_id']
        client_secret = self.connector_information['client_secret']
        site_id = self.connector_information['site_id']

        #Auth
        token_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
        token_data = {
            'grant_type': 'client_credentials',
            'client_id': client_id,
            'client_secret': client_secret,
            'scope': 'https://graph.microsoft.com/.default'
        }
        token_r = requests.post(token_url, headers={"Content-Type":"application/x-www-form-urlencoded"}, data=token_data)
        token = token_r.json().get('access_token')

        # Get all document libraries in the site
        headers = {
            'Authorization': f'Bearer {token}',
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        doc_libraries_url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives"
        doc_libraries_response = requests.get(doc_libraries_url, headers=headers)
        doc_libraries_info = doc_libraries_response.json()

        # Iterate through all document libraries
        for doc_library in doc_libraries_info.get('value', []):
            doc_library_id = doc_library['id']
            yield from self.process_folder_delta(site_id=site_id, drive_id=doc_library_id, folder_id='root', headers=headers, last_run=last_run)
        
    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        """Connect to source and download file into local storage"""
        #Download file
        id = cloudFile.id
        file_identifier = cloudFile.file_identifier
        type = cloudFile.type
        metadata = cloudFile.metadata
        import requests
        file_r = requests.get(file_identifier)
        with tempfile.NamedTemporaryFile(delete=False, suffix=f"_{id}") as temp_file:
            temp_file.write(file_r.content)
        yield LocalFile(file_path=temp_file.name, metadata=metadata, id=id, type=type)
    
    def validate(self) -> bool:
        try:
            tenant_id = self.connector_information['tenant_id']
            client_id = self.connector_information['client_id']
            client_secret = self.connector_information['client_secret']
            site_id = self.connector_information['site_id']
        except:
            raise ValueError(f"Required properties not set. Required properties: {self.required_properties}")
        
        if not all(x in self.available_metadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")
        
        try:
            token_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
            token_data = {
                'grant_type': 'client_credentials',
                'client_id': client_id,
                'client_secret': client_secret,
                'scope': 'https://graph.microsoft.com/.default'
            }
            token_r = requests.post(token_url, headers={"Content-Type":"application/x-www-form-urlencoded"}, data=token_data)
            token = token_r.json().get('access_token')

            # Get all document libraries in the site
            headers = {
                'Authorization': f'Bearer {token}',
                'Accept': 'application/json',
                'Content-Type': 'application/json'
            }
            doc_libraries_url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives"
            requests.get(doc_libraries_url, headers=headers)
        except Exception as e:
            raise SharepointConnectionException(f"Connection to Sharepoint failed, check credentials. See Exception: {e}")       
        return True 


from Loaders import (
    AutoLoader,
    LoaderEnum,
    MarkdownLoader,
    NeumJSONLoader,
    NeumCSVLoader,
    PDFLoader,
    HTMLLoader,
)
from neumai.Shared.Selector import Selector

def as_loader(dct:dict):
    if dct == None:
        return AutoLoader()
    loader_name = dct.get("loader_name", None)
    loader_information = dct.get("loader_information", None)
    selector = Selector.as_selector(dct.get("selector", None))
    if loader_name == LoaderEnum.autoloader:
        return AutoLoader(loader_information=loader_information, selector=selector)
    elif loader_name == LoaderEnum.htmlloader:
        return HTMLLoader(loader_information=loader_information, selector=selector)
    elif loader_name == LoaderEnum.markdownloader:
        return MarkdownLoader(loader_information=loader_information, selector=selector)
    elif loader_name == LoaderEnum.neumjsonloader:
        return NeumJSONLoader(loader_information=loader_information, selector=selector)
    elif loader_name == LoaderEnum.neumcsvloader:
        return NeumCSVLoader(loader_information=loader_information, selector=selector)
    elif loader_name == LoaderEnum.pdfloader:
        return PDFLoader(loader_information=loader_information, selector=selector)
    else: 
        return AutoLoader(loader_information=loader_information, selector=selector)


from typing import List, Generator
from Shared.NeumDocument import NeumDocument
from Shared.LocalFile import LocalFile
from Loaders.Loader import Loader
from langchain.document_loaders import UnstructuredHTMLLoader

class HTMLLoader(Loader):
    """" HTML Loader \n loader_information requires: [ ]"""

    @property
    def loader_name(self) -> str:
        return "HTMLLoader"
    
    @property
    def required_properties(self) -> List[str]:
        return []

    @property
    def optional_properties(self) -> List[str]:
        return []
    
    @property
    def available_metadata(self) -> List[str]:
        return []

    @property
    def availableContent(self) -> List[str]:
        return []

    def load(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        """Load data into Document objects."""
        loader = UnstructuredHTMLLoader(file_path=file.file_path)
        documents = loader.load()
        # join the file and document metadata objects
        for doc in documents:
            yield NeumDocument(id=file.id, content=doc.page_content, metadata=file.metadata)
    
    def validate(self) -> bool:
        return True   


from .CloudFile import CloudFile
from .LocalFile import LocalFile
from .NeumDocument import NeumDocument
from .NeumVector import NeumVector
from .Selector import Selector
from .NeumSinkInfo import NeumSinkInfo
from .NeumSearch import NeumSearchResult
from .Exceptions import (
    AzureBlobConnectionException,
    CloudFileEmptyException,
    InvalidDataConnectorException,
    InvalidSinkConnectorException,
    LocalFileEmptyException,
    NeumDocumentEmptyException,
    NeumSearchResultEmptyException,
    NeumSinkInfoEmptyException,
    OpenAIConnectionException,
    PostgresConnectionException,
    PineconeConnectionException,
    PineconeInsertionException,
    PineconeIndexInfoException,
    PineconeQueryException,
    QdrantInsertionException,
    QdrantIndexInfoException,
    QdrantQueryException,
    S3ConnectionException,
    SharepointConnectionException,
    SinglestoreConnectionException,
    SinglestoreInsertionException,
    SinglestoreIndexInfoException,
    SinglestoreQueryException,
    SourceConnectorEmptyException,
    SupabaseConnectionException,
    SupabaseInsertionException,
    SupabaseIndexInfoException,
    SupabaseQueryException,
    WeaviateConnectionException,
    WeaviateInsertionException,
    WeaviateIndexInfoException,
    WeaviateQueryException,
    WebsiteConnectionException,
)


from abc import abstractmethod, ABC
from typing import List, Generator
from Shared.LocalFile import LocalFile
from Shared.CloudFile import CloudFile
from Shared.Selector import Selector
from datetime import datetime

class DataConnector(ABC):
    def __init__(self, connector_information:dict = {}, selector:Selector = Selector(to_embed=[], to_metadata=[])) -> None:
        self.connector_information = connector_information
        self.selector = selector

    @property
    @abstractmethod
    def connector_name(self) -> str:
        pass

    @property
    @abstractmethod
    def required_properties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def optional_properties(self) -> List[str]:
        pass
    
    @property
    @abstractmethod
    def available_metadata(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def schedule_avaialable(self) -> bool:
        pass

    @property
    @abstractmethod
    def auto_sync_available(self) -> bool:
        pass

    @property
    @abstractmethod
    def compatible_loaders(self) -> List[str]:
        pass

    @abstractmethod
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        """Connect to source and download file into local storage"""
    
    @abstractmethod
    def connect_and_list_delta(self, last_run:datetime) -> Generator[LocalFile, None, None]:
        """Check for changes in the source"""
        """Code to be pushed to a worker and run on a schedule"""
    
    @abstractmethod
    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        """Connect to source and download file into local storage"""

    @abstractmethod
    def validate(self) -> bool:
        """Validate if the connector is correctly configured"""

    # To do auto_sync logic.

    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['connector_name'] = self.connector_name
        json_to_return['connector_information'] = self.connector_information
        json_to_return['selector'] = self.selector.toJson()
        return json_to_return
    
    def to_model(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..
        This is different than toJson, here we use it to create a model, we don't want to return the api key in the body back. Eventualyl this should be its own class...
        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['connector_name'] = self.connector_name
        json_to_return['connector_information'] = self.connector_information
        json_to_return['selector'] = self.selector.to_model()
        return json_to_return

    def config(self):
        json_to_return = {}
        json_to_return['required_properties'] = self.required_properties
        json_to_return['optional_properties'] = self.optional_properties
        json_to_return['available_metadata'] = self.available_metadata
        json_to_return['compatible_loaders'] = self.compatible_loaders
        json_to_return['schedule_avaialable'] = self.schedule_avaialable
        json_to_return['auto_sync_available'] = self.auto_sync_available
        return json_to_return


from typing import List, Generator
from datetime import datetime
from abc import ABC

from DataConnectors.DataConnector import DataConnector
from DataConnectors.ConnectorHelper import as_connector
from Chunkers.Chunker import Chunker
from Chunkers.RecursiveChunker import RecursiveChunker
from Chunkers.ChunkerHelper import as_chunker
from Loaders.Loader import Loader
from Loaders.AutoLoader import AutoLoader
from Loaders.LoaderHelper import as_loader
from Shared.NeumDocument import NeumDocument
from Shared.LocalFile import LocalFile
from Shared.CloudFile import CloudFile
from Shared.Exceptions import SourceConnectorEmptyException

class SourceConnector(ABC):
    def __init__(self, connector:DataConnector, chunker:Chunker = RecursiveChunker(), loader:Loader = AutoLoader(), customMetadata:dict = {}):
        self.connector = connector
        self.chunker = chunker
        self.loader = loader
        self.customMetadata = customMetadata

    def list_files_full(self) -> Generator[CloudFile, None, None]:
        yield from self.connector.connect_and_list_full()

    def list_files_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        yield from self.connector.connect_and_list_delta(last_run=last_run)

    def download_files(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        yield from self.connector.connect_and_download(cloudFile=cloudFile)

    def load_data(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        yield from self.loader.load(file=file)

    def chunk_data(self, document:NeumDocument) -> Generator[List[NeumDocument], None, None]:
        for chunk_set in self.chunker.chunk(documents=[document]):
            chunk_set_with_custom_metadata = [NeumDocument(id=chunk.id, content=chunk.content, metadata={**chunk.metadata, **self.customMetadata, **{"text":chunk.content}}) for chunk in chunk_set]
            yield chunk_set_with_custom_metadata
    
    def validation(self) -> bool:
        core_validation = self.connector.validate() and self.loader.validate() and self.chunker.validate()
        loader_validation = self.loader.loader_name in self.connector.compatible_loaders
        return core_validation and loader_validation

    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['customMetadata'] = self.customMetadata
        json_to_return['connector'] = self.connector.toJson()
        json_to_return['chunker'] = self.chunker.toJson()
        json_to_return['loader'] = self.loader.toJson()
        return json_to_return
    
    def to_model(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..
        This is different han toJson, here we use it to create a model, we don't want to return the api key in the body back. Eventualyl this should be its own class...
        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['customMetadata'] = self.customMetadata
        json_to_return['connector'] = self.connector.to_model()
        json_to_return['chunker'] = self.chunker.to_model()
        json_to_return['loader'] = self.loader.to_model()
        return json_to_return

    def as_source_connector(dct:dict):
        if dct == None:
            raise SourceConnectorEmptyException("Received empty dict when converting to as_source_connector")
        return SourceConnector(
            customMetadata=dct.get('customMetadata', {}),
            connector=as_connector(dct.get("connector", None)),
            chunker=as_chunker(dct.get("chunker", None)),
            loader=as_loader(dct.get("loader", None))
        )



from enum import Enum

class DataConnectorEnum(str, Enum):
    azureblobconnector = "azureblobconnector"
    neumfileconnector = "neumfileconnector"
    neumwebsiteconnector = "neumwebsiteconnector"
    postgresconnector = "postgresconnector"
    s3connector = "s3connector"
    sharepointconnector = "sharepointconnector"
    singlestoreconnector = "singlestoreconnector"
    supabasestorageconnector = "supabasestorageconnector"

    def as_data_connector_enum(data_connector_name: str):
        if data_connector_name == None or data_connector_name == "":
            return None
        try:
            enum_to_return = DataConnectorEnum[data_connector_name.lower()]
            return enum_to_return
        except KeyError as e:
            return None


from enum import Enum

class SinkConnectorEnum(str, Enum):
    pineconesink = "pineconesink"
    qdrantsink = "qdrantsink"
    singlestoresink = "singlestoresink"
    supabasesink = "supabasesink"
    weaviatesink = "weaviatesink"

    def as_data_connector_enum(sink_connector_name: str):
        if sink_connector_name == None or sink_connector_name == "":
            return None
        try:
            enum_to_return = SinkConnectorEnum[sink_connector_name.lower()]
            return enum_to_return
        except KeyError as e:
            return None


from datetime import datetime
from DataConnectors import DataConnector
from typing import List, Generator
from azure.storage.blob import BlobClient, ContainerClient
from Shared.LocalFile import LocalFile
from Shared.CloudFile import CloudFile
from Shared.Exceptions import AzureBlobConnectionException
import tempfile
import os

class AzureBlobConnector(DataConnector):
    """" Neum File Connector """

    @property
    def connector_name(self) -> str:
        return "AzureBlobConnector"
    
    @property
    def required_properties(self) -> List[str]:
        return ["connection_string", "container_name"]

    @property
    def optional_properties(self) -> List[str]:
        return []
    
    @property
    def available_metadata(self) -> str:
        return ['name', 'last_modified', 'creation_time', 'last_access_on']

    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["AutoLoader", "HTMLLoader", "MarkdownLoader", "NeumCSVLoader", "NeumJSONLoader", "PDFLoader"]
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        # Connect to Azure Blob Storage
        connection_string = self.connector_information['connection_string']
        container_name = self.connector_information['container_name']
        container = ContainerClient.from_connection_string(
            conn_str=connection_string, container_name=container_name
        )

        #Process files
        file_list = container.list_blobs()
        for file in file_list:
            name = file.name
            metadata = {
                "creation_time": file.creation_time.isoformat(),
                "last_modified": file.last_modified.isoformat(),
                "last_access_on": file.last_accessed_on.isoformat() if file.last_accessed_on is not None else None
            }
            selected_metadata  = {k: metadata[k] for k in self.selector.to_metadata if k in metadata}
            yield CloudFile(file_identifier=name, metadata=selected_metadata, id = name)

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # Connect to Azure Blob Storage
        connection_string = self.connector_information['connection_string']
        container_name = self.connector_information['container_name']
        container = ContainerClient.from_connection_string(
            conn_str=connection_string, container_name=container_name
        )

        #Process files
        file_list = container.list_blobs()
        for file in file_list:
            last_update_date = file.last_modified
            if(last_run < last_update_date):
                name = file.name
                metadata = {
                    "creation_time": file.creation_time.isoformat(),
                    "last_modified": file.last_modified.isoformat(),
                    "last_access_on": file.last_accessed_on.isoformat() if file.last_accessed_on is not None else None
                }
                selected_metadata  = {k: metadata[k] for k in self.selector.to_metadata if k in metadata}
                yield CloudFile(file_identifier=name, metadata=selected_metadata, id=name)

    def connect_and_download(self,  cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        # Connect to Azure Blob Storage
        connection_string = self.connector_information['connection_string']
        container_name = self.connector_information['container_name']
        
        client = BlobClient.from_connection_string(conn_str=connection_string, container_name=container_name, blob_name=cloudFile.file_identifier)
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = f"{temp_dir}/{container_name}/{cloudFile.file_identifier}"
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(f"{file_path}", "wb") as file:
                blob_data = client.download_blob()
                blob_data.readinto(file)
            yield LocalFile(file_path=file_path, metadata=cloudFile.metadata, id=cloudFile.id)
        
    def validate(self) -> bool:
        try:
            connection_string = self.connector_information['connection_string']
            container_name = self.connector_information['container_name']
        except:
            raise ValueError(f"Required properties not set. Required properties: {self.required_properties}")
        
        if not all(x in self.available_metadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")

        try:
            ContainerClient.from_connection_string(
                conn_str=connection_string, container_name=container_name
            )
        except Exception as e:
            raise AzureBlobConnectionException(f"Connection to Azure Blob Storage failed, check credentials. See Exception: {e}")
        return True     


    
 


class AzureBlobConnectionException(Exception):
    """Raised if establishing a connection to Azure Blob fails"""
    pass

class CloudFileEmptyException(Exception):
    """Raised when the cloud file dictionary is empty"""
    pass

class LocalFileEmptyException(Exception):
    """Raised when the local file dictionary is empty"""
    pass

class InvalidDataConnectorException(Exception):
    """Raised when an invalid data connector is detected"""
    pass

class InvalidEmbedConnectorException(Exception):
    """Raised when an invalid embed connector is detected"""
    pass

class InvalidSinkConnectorException(Exception):
    """Raised when an invalid sink connector is detected"""
    pass

class NeumDocumentEmptyException(Exception):
    """Raised when the Neum document dictionary is empty"""
    pass

class NeumSearchResultEmptyException(Exception):
    """Raised when the Neum search result dictionary is empty"""
    pass

class NeumSinkInfoEmptyException(Exception):
    """Raised when the Neum search result dictionary is empty"""
    pass

class OpenAIConnectionException(Exception):
    """Raised if establishing a connection to OpenAI fails"""
    pass

class PineconeConnectionException(Exception):
    """Raised if establishing a connection to Pinecone fails"""
    pass

class PineconeInsertionException(Exception):
    """Raised if inserting into Pinecone fails"""
    pass

class PineconeIndexInfoException(Exception):
    """Raised if getting index info from Pinecone fails"""
    pass

class PineconeQueryException(Exception):
    """Raised if querying Pinecone fails"""
    pass

class QdrantInsertionException(Exception):
    """Raised if inserting into Qdrant fails"""
    pass

class QdrantIndexInfoException(Exception):
    """Raised if getting index info from Qdrant fails"""
    pass

class QdrantQueryException(Exception):
    """Raised if querying Qdrant fails"""
    pass

class PostgresConnectionException(Exception):
    """Raised if establishing a connection to a Postgres db fails"""
    pass

class S3ConnectionException(Exception):
    """Raised if establishing a connection to AWS S3 fails"""
    pass

class SharepointConnectionException(Exception):
    """Raised if establishing a connection to Sharepoint fails"""
    pass

class SinglestoreConnectionException(Exception):
    """Raised if establishing a connection to Singlestore fails"""
    pass

class SinglestoreInsertionException(Exception):
    """Raised if inserting into Singlestore fails"""
    pass

class SinglestoreIndexInfoException(Exception):
    """Raised if getting index info from Singlestore fails"""
    pass

class SinglestoreQueryException(Exception):
    """Raised if querying Singlestore fails"""
    pass

class SourceConnectorEmptyException(Exception):
    """Raised when the SourceConnector dictionary is empty"""
    pass

class SupabaseConnectionException(Exception):
    """Raised if establishing a connection to Supabase fails"""
    pass

class SupabaseInsertionException(Exception):
    """Raised if inserting into Supabase fails"""
    pass

class SupabaseIndexInfoException(Exception):
    """Raised if getting index info from Supabase fails"""
    pass

class SupabaseQueryException(Exception):
    """Raised if querying Supabase fails"""
    pass

class WeaviateConnectionException(Exception):
    """Raised if establishing a connection to Weaviate fails"""
    pass

class WeaviateInsertionException(Exception):
    """Raised if inserting into Weaviate fails"""
    pass

class WeaviateIndexInfoException(Exception):
    """Raised if getting index info from Weaviate fails"""
    pass

class WeaviateQueryException(Exception):
    """Raised if querying Weaviate fails"""
    pass

class WebsiteConnectionException(Exception):
    """Raised if establishing a connection to a website fails"""
    pass



from abc import ABC
from Shared.Exceptions import NeumSearchResultEmptyException

class NeumSearchResult(ABC):
    def __init__(self, id:str, metadata:dict, score:float) -> None:
        self.id =  id
        self.metadata = metadata
        self.score = score

    def as_search_result(dct:dict):
        if dct == None:
            raise NeumSearchResultEmptyException("Received empty dict when converting to as_search_result")
        return NeumSearchResult(
            id=dct.get("id", None),
            metadata=dct.get("metadata", None),
            score=dct.get("score", None),
        )
    
    def toJson(self) -> dict:
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['id'] = self.id
        json_to_return['metadata'] = self.metadata
        json_to_return['score'] = self.score
        return json_to_return


from abc import abstractmethod, ABC
from neumai.Shared.NeumDocument import NeumDocument
from typing import List, Generator

class Chunker(ABC):
    def __init__(self, chunker_information:dict = {}) -> None:
        self.chunker_information = chunker_information
    
    @property
    @abstractmethod
    def chunker_name(self) -> str:
        pass

    @property
    @abstractmethod
    def required_properties(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def optional_properties(self) -> List[str]:
        pass

    @abstractmethod
    def chunk(self, documents:List[NeumDocument]) -> Generator[List[NeumDocument], None, None]:
        """Chunk documents into more documents"""

    @abstractmethod
    def validate(self) -> bool:
        """Validate if the chunker is correctly configured"""

    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['chunker_name'] = self.chunker_name
        json_to_return['chunker_information'] = self.chunker_information
        return json_to_return
    
    def to_model(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..
        This is different han toJson, here we use it to create a model, we don't want to return the api key in the body back. Eventualyl this should be its own class...
        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['chunker_name'] = self.chunker_name
        json_to_return['chunker_information'] = self.chunker_information
        return json_to_return

    def config(self):
        json_to_return = {}
        json_to_return['required_properties'] = self.required_properties
        json_to_return['optional_properties'] = self.optional_properties
        return json_to_return


from typing import List, Optional, Tuple
from SinkConnectors.SinkConnector import SinkConnector
from Shared.NeumVector  import NeumVector
from Shared.NeumSinkInfo import NeumSinkInfo
from Shared.NeumSearch import NeumSearchResult
from Shared.Exceptions import(
    WeaviateConnectionException,
    WeaviateInsertionException,
    WeaviateIndexInfoException,
    WeaviateQueryException
)
import weaviate
from weaviate.util import generate_uuid5, _capitalize_first_letter

class WeaviateSink(SinkConnector):
    """ Weaviate Sink\n
    sink_information requires : [ url , api_key ]"""

    @property
    def sink_name(self) -> str:
        return 'WeaviateSink'
    
    @property
    def required_properties(self) -> List[str]:
        return ['url', 'api_key']

    @property
    def optional_properties(self) -> List[str]:
        return ['class_name', 'num_workers', 'shard_count', 'batch_size', 'is_dynamic_batch', 'batch_connection_error_retries']

    def validation(self) -> bool:
        """Validate connector setup"""
        try:
            url = self.sink_information["url"]
            if 'https' not in url:
                api_key = self.sink_information["api_key"]
        except:
            raise ValueError(f"Required properties not set. Required properties: {self.required_properties}")
        try:
            if 'https' not in url:
                client = weaviate.Client(
                    url=url
                )
            else:
                api_key = self.sink_information["api_key"]
                client = weaviate.Client(
                    url=url,
                    auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
                )
        except Exception as e:
            raise WeaviateConnectionException(f"Weaviate couldn't be initialized. See exception: {e}")
        return True 

    def _check_batch_result(self, results: Optional[List[dict[str, any]]], task_id: str, partial_failure: dict):
        if results is not None:
            for result in results:
                if "result" in result and "errors" in result["result"]:
                    if "error" in result["result"]["errors"]:
                        print(f"[ERROR] Task Id {task_id} encountered an error when batching to weaviate {result['result']}")
                        partial_failure['did_fail'] = True
                        partial_failure['latest_failure'] = result["result"]["errors"]["error"]
                        partial_failure['number_of_failures'] += 1

    def store(self, pipeline_id: str, vectors_to_store:List[NeumVector], task_id:str = "") -> Tuple[List, dict]:
        url = self.sink_information["url"]
        num_workers = self.sink_information.get('num_workers', 1)
        shard_count = self.sink_information.get('shard_count', 1)
        batch_size = self.sink_information.get('batch_size', 100)
        is_dynamic_batch = self.sink_information.get('is_dynamic_batch', False)
        batch_connection_error_retries = self.sink_information.get('batch_connection_error_retries', 3)
        class_name = self.sink_information.get('class_name', f"pipeline_{pipeline_id.replace('-','_')}")
        partial_failure = {'did_fail': False, 'latest_failure': None, 'number_of_failures': 0}

        if 'https' not in url:
            client = weaviate.Client(
                url=url
            )
        else:
            api_key = self.sink_information["api_key"]
            client = weaviate.Client(
                url=url,
                auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
            )
        try:
            client.schema.create_class({
                "class": class_name,
                "shardingConfig":{"desiredCount":shard_count},
            })
        except weaviate.UnexpectedStatusCodeException as e:
            if 'already exists' not in e.message:
                raise WeaviateInsertionException(f"Error when creating class in weaviate. Error: {e}")
            
        with client.batch.configure(
            batch_size=batch_size,
            callback=lambda results: self._check_batch_result(results, task_id, partial_failure),
            num_workers=num_workers,
            dynamic=is_dynamic_batch,
            connection_error_retries=batch_connection_error_retries
        ) as batch:
            for i in range(0, len(vectors_to_store)):
                try:
                    batch.add_data_object(
                        data_object=vectors_to_store[i].metadata,
                        class_name=class_name,
                        vector=vectors_to_store[i].vector,
                        uuid=generate_uuid5(vectors_to_store[i].id)
                    )
                except Exception as e:
                    print(f"[ERROR] Got exception from Weaviate when adding data object to batch for task {task_id} and batch # {i}. Exception: {str(e)}")
                    raise WeaviateInsertionException(f"Error when adding data object to Weaviate. Error: {e}")

        return len(vectors_to_store)

    def search(self, vector: List[float], number_of_results: int, pipeline_id: str) -> List[NeumSearchResult]:
        api_key = self.sink_information["api_key"]
        url = self.sink_information['url']
        # Weaviate requires first letter to be capitalized
        class_name = _capitalize_first_letter(self.sink_information.get('class_name', f"Pipeline_{pipeline_id.replace('-', '_')}"))
        client = weaviate.Client(
            url=url,
            auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
        )
        
        try:
            class_schema = client.schema.get(class_name)
        except Exception as e:
            raise WeaviateQueryException(f"There was an error retrieving the class schema from weaviate")

        full_class_schema_properties = [property['name'] for property in class_schema['properties']]
        matches = []
        try:
            search_result = (
                client.query
                .get(class_name=class_name, properties=full_class_schema_properties)
                .with_near_vector(content = {
                    'vector' : vector
                })
                .with_limit(number_of_results)
                .with_additional(['id','certainty'])
                .do()
            )

            for result in search_result["data"]["Get"][class_name]:
                # unify our api with the metadata.. or just return whatever metadata we have. (?)
                matches.append(NeumSearchResult(id=result['_additional']['id'], score=result['_additional']['certainty'], metadata= {k: v for k, v in result.items() if k != "_additional"}))
        except Exception as e:
            raise WeaviateQueryException(f"There was an error querying weaviate. Error {e}")
        return matches

    def info(self, pipeline_id:str) -> NeumSinkInfo:
        api_key = self.sink_information["api_key"]
        url = self.sink_information['url']
        
        class_name = _capitalize_first_letter(self.sink_information.get('class_name', f"Pipeline_{pipeline_id.replace('-', '_')}"))
        client = weaviate.Client(
            url=url,
            auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
        )
        try:
            response = (
                client.query
                .aggregate(class_name=class_name)
                .with_meta_count()
                .do()
            )
            vectors_stored_in_class = response["data"]["Aggregate"][class_name]["meta"]["count"]
            return NeumSinkInfo(number_vectors_stored=vectors_stored_in_class)
        except Exception as e:
            raise WeaviateIndexInfoException(f"There was an error getting class info from weaviate {e}")


from Chunkers import (
    CharacterChunker,
    ChunkerEnum,
    CustomChunker,
    RecursiveChunker
)    

def as_chunker(dct:dict):
    if dct == None:
        return RecursiveChunker()
    chunker_name = dct.get("chunker_name", None)
    chunker_information = dct.get("chunker_information", None)
    if chunker_name == ChunkerEnum.characterchunker:
        return CharacterChunker(chunker_information=chunker_information)
    elif chunker_name == ChunkerEnum.customchunker:
        return CustomChunker(chunker_information=chunker_information)
    elif chunker_name == ChunkerEnum.recursivechunker:
        return RecursiveChunker(chunker_information=chunker_information)
    else:
        return RecursiveChunker(chunker_information=chunker_information)


from typing import List, Tuple
from EmbedConnectors.EmbedConnector import EmbedConnector
from Shared.NeumDocument import NeumDocument
from langchain.embeddings import azure_openai

class AzureOpenAIEmbed(EmbedConnector):
    @property
    def embed_name(self) -> str:
        return 'AzureOpenAIEmbed'
    
    @property
    def required_properties(self) -> List[str]:
        return ['deployment_name', 'api_key', 'endpoint']

    @property
    def optional_properties(self) -> List[str]:
        return []

    def validation(self) -> bool:
        """Validate connector setup"""
        try:
            deployment_name = self.embed_information["deployment_name"]
            api_key = self.embed_information["api_key"]
            api_base = self.embed_information["endpoint"]
        except Exception as e:
            raise ValueError(f"Required properties not set. Required properties: {self.required_properties}")
        return True 

    def embed(self, documents:List[NeumDocument]) -> Tuple[List, dict]:
        """Generate embeddings with Azure OpenAI"""
        max_retries = self.embed_information.get('max_retries', 20)
        chunk_size = max(self.embed_information.get('chunk_size', 16), 16)

        embedding = azure_openai.AzureOpenAIEmbeddings(
            max_retries=max_retries,
            chunk_size=chunk_size,
            azure_deployment=self.embed_information["deployment_name"],
            api_key=self.embed_information['api_key'],
            azure_endpoint=self.embed_information['endpoint'],
        )
        embeddings = []
        texts = [x.content for x in documents]
        embeddings  = embedding.embed_documents(texts=texts)
        info = {
            "estimated_cost":str("Not implemented"),
            "total_tokens":str("Not implemented"),
            "attempts_used":str("Not implemented")
        }
        return embeddings,info
    
    def embed_query(self, query: str) -> List[float]:
        """Generate embeddings for a single query using Azure OpenAI"""

        embedding = azure_openai.AzureOpenAIEmbeddings(
            azure_deployment=self.embed_information["deployment_name"],
            api_key=self.embed_information['api_key'],
            azure_endpoint=self.embed_information['endpoint'],
        )
        return embedding.embed_query(query)


from SinkConnectors import (
    PineconeSink,
    WeaviateSink,
    QdrantSink,
    SingleStoreSink,
    SinkConnectorEnum,
    SupabaseSink
)

from Shared.Exceptions import InvalidSinkConnectorException

available_connectors = [enum.value for enum in list(SinkConnectorEnum)]

def as_sink(dct:dict):
    if dct == None:
        raise InvalidSinkConnectorException("Must supply a data connector configuration")
    if not isinstance(dct, dict):
        raise InvalidSinkConnectorException("Data connector configuration needs to be a dictionary")
    sink_name = dct.get("sink_name", None)
    sink_information = dct.get("sink_information", None)
    if sink_name == SinkConnectorEnum.pineconesink:
        return PineconeSink(sink_information=sink_information)
    elif sink_name == SinkConnectorEnum.qdrantsink:
        return QdrantSink(sink_information=sink_information)
    elif sink_name == SinkConnectorEnum.singlestoresink:
        return SingleStoreSink(sink_information=sink_information)
    elif sink_name == SinkConnectorEnum.supabasesink:
        return SupabaseSink(sink_information=sink_information)
    elif sink_name == SinkConnectorEnum.weaviatesink:
        return WeaviateSink(sink_information=sink_information)
    else:
        raise InvalidSinkConnectorException(f"{sink_name} is an invalid sink connector. Available connectors: {available_sink_connectors}]")



from typing import List
from Shared.NeumSearch import NeumSearchResult
from Shared.NeumSinkInfo import NeumSinkInfo
from SinkConnectors.SinkConnector import SinkConnector
from Shared.NeumVector  import NeumVector
from Shared.Exceptions import (
    PineconeConnectionException,
    PineconeInsertionException,
    PineconeIndexInfoException,
    PineconeQueryException,
)
import pinecone

class  PineconeSink(SinkConnector):
    """ Pinecone Sink\n
    sink_information requires : [ 'api_key', 'environment', 'index' ]"""
        
    @property
    def sink_name(self) -> str:
        return 'PineconeSink'
    
    @property
    def required_properties(self) -> List[str]:
        return ['api_key', 'environment', 'index']

    @property
    def optional_properties(self) -> List[str]:
        return ['namespace']

    def validation(self) -> bool:
        """Validate connector setup"""
        import pinecone
        try:
            api_key = self.sink_information['api_key']
            environment = self.sink_information['environment']
            index = self.sink_information['index']
        except:
            raise ValueError(f"Required properties not set. Required properties: {self.required_properties}")
        try:
            pinecone.init(api_key=api_key, environment=environment)    
            index = pinecone.Index(index_name=index)
            index.describe_index_stats()
        except Exception as e:
            raise PineconeConnectionException(f"Pinecone connection couldn't be initialized. See exception: {e}")
        return True 

    def store(self, pipeline_id: str, vectors_to_store:List[NeumVector], task_id:str = "") -> int:
        api_key =  self.sink_information['api_key']
        environment = self.sink_information['environment']
        index = self.sink_information['index']
        namespace = self.sink_information.get("namespace", f"pipeline_{pipeline_id}")
        try:
            pinecone.init(api_key=api_key, environment=environment)    
            index = pinecone.Index(index_name=index)
            batch_size = 32 # how does changing this affect ?
            vectors_stored = 0
            for i in range(0, len(vectors_to_store), batch_size):
                # set end position of batch
                i_end = min(i + batch_size, len(vectors_to_store))
                # get batch of texts and ids
                vector_batch = vectors_to_store[i:i_end]
                to_upsert = [(vector.id, vector.vector, vector.metadata) for vector in vector_batch]
                result = index.upsert(vectors=to_upsert, namespace=namespace)
                vectors_stored += result['upserted_count'] 
        except Exception as e:
            raise PineconeInsertionException(f"Failed to store in Pinecone. Exception - {e}")
        return int(vectors_stored)
    
    def search(self, vector: List[float], number_of_results:int, pipeline_id:str) -> List[NeumSearchResult]:
        import pinecone
        api_key = self.sink_information["api_key"]
        environment = self.sink_information['environment']
        index = self.sink_information['index']
        namespace = self.sink_information.get("namespace", f"pipeline_{pipeline_id}")
        try:
            pinecone.init(      
                api_key=api_key,      
                environment=environment)    
            index = pinecone.Index(index)
            results = index.query(vector=vector, top_k=number_of_results, namespace=namespace, include_values=False, include_metadata=True)["matches"]
        except Exception as e:
            raise PineconeQueryException(f"Failed to query pinecone. Exception - {e}")
        
        matches = []
        for result in results:
            matches.append(NeumSearchResult(id= result["id"], metadata=result["metadata"], score=result["score"]))
        return matches
    
    def info(self, pipeline_id: str) -> NeumSinkInfo:
        import pinecone
        api_key = self.sink_information["api_key"]
        environment = self.sink_information['environment']
        index = self.sink_information['index']
        namespace = self.sink_information.get("namespace", f"pipeline_{pipeline_id}")
        try:
            pinecone.init(      
                api_key=api_key,      
                environment=environment)    
            index = pinecone.Index(index)
            namespaces = index.describe_index_stats()["namespace"]
            if namespace in namespaces:
                return NeumSinkInfo(number_vectors_stored=namespaces[namespace]["vector_count"])
        except Exception as e:
            raise PineconeIndexInfoException(f"Failed to get info for pinecone. Exception - {e}")


from typing import List, Generator
from Shared.NeumDocument import NeumDocument
from Shared.LocalFile import LocalFile
from Loaders.Loader import Loader
from Loaders.NeumCSVLoader import NeumCSVLoader
from Loaders.NeumJSONLoader import NeumJSONLoader
from Loaders.HTMLLoader import HTMLLoader
from Loaders.MarkdownLoader import MarkdownLoader
from Loaders.PDFLoader import PDFLoader
from langchain.document_loaders import UnstructuredFileLoader

class AutoLoader(Loader):
    """ Auto Loader\n loader_information contains: [] """

    @property
    def loader_name(self) -> str:
        return "AutoLoader"
    
    @property
    def required_properties(self) -> List[str]:
        return []

    @property
    def optional_properties(self) -> List[str]:
        return []
    
    @property
    def available_metadata(self) -> List[str]:
        return []

    @property
    def availableContent(self) -> List[str]:
        return []

    def load(self, file:LocalFile) -> Generator[NeumDocument, None, None]:
        """Load data into Document objects."""
        if "csv" in file.type:
            loader = NeumCSVLoader()
        elif "string" in file.type:
            yield NeumDocument(content=file.in_mem_data, metadata=file.metadata, id=file.id)
            return
        elif "application/octet-stream" in file.type:
            loader = MarkdownLoader()
        elif "html" in file.type:
            loader = HTMLLoader()
        elif "pdf" in file.type:
            loader = PDFLoader()
        elif "json" in file.type:
            loader = NeumJSONLoader()
        else:
            loader = UnstructuredFileLoader(file_path=file.file_path)
            documents = loader.load()
            for doc in documents:
                yield NeumDocument(id=file.id, content=doc.page_content, metadata=file.metadata)
            return 
        
        yield from loader.load(file=LocalFile)
    
    def validate(self) -> bool:
        return True   


from neumai.EmbedConnectors import (
    OpenAIEmbed,
    ReplicateEmbed,
    AzureOpenAIEmbed
)

from EmbedConnectors.EmbedConnectorEnum import EmbedConnectorEnum
from Shared.Exceptions import InvalidEmbedConnectorException

available_embed_connectors = [enum.value for enum in list(EmbedConnectorEnum)]

def as_embed(dct:dict):
    if dct == None:
        raise InvalidEmbedConnectorException("Must supply an embed connector configuration")
    if not isinstance(dct, dict):
        raise InvalidEmbedConnectorException("Embed connector configuration needs to be a dictionary")
    
    embed_connector_name = dct.get("embed_name", None)
    embed_connector_enum = EmbedConnectorEnum.as_embed_connector_enum(embed_connector_name=embed_connector_name)
    embed_information = dct.get("embed_information", None)
    if embed_connector_enum == EmbedConnectorEnum.azureopenaiembed:
        return AzureOpenAIEmbed(embed_information=embed_information)
    elif embed_connector_enum == EmbedConnectorEnum.openaiembed:
        return OpenAIEmbed(embed_information=embed_information)
    elif embed_connector_enum == EmbedConnectorEnum.replicateembed:
        return ReplicateEmbed(embed_information=embed_information)
    else:
        raise InvalidEmbedConnectorException(f"{embed_connector_name} is an invalid embed connector. Available connectors: {available_embed_connectors}] ")


from typing import List, Generator
from Shared.NeumDocument import NeumDocument
from Shared.LocalFile import LocalFile
from Loaders.Loader import Loader
import csv

class NeumCSVLoader(Loader):
    """" Neum CSV Loader """

    @property
    def loader_name(self) -> str:
        return "NeumCSVLoader"
    
    @property
    def required_properties(self) -> List[str]:
        return []

    @property
    def optional_properties(self) -> List[str]:
        return ["id_key" , "source_column" , "encoding", "csv_args"]
    
    @property
    def available_metadata(self) -> List[str]:
        return ["custom"]

    @property
    def availableContent(self) -> List[str]:
        return ["custom"]
    
    def validate(self) -> bool:
        return True   

    def load(self, file: LocalFile) -> Generator[NeumDocument, None, None]:
        source_column = self.loader_information.get('source_coulmn', None)
        encoding = self.loader_information.get('encoding', "utf-8-sig") # modify to use encoding
        csv_args = self.loader_information.get('csv_args', None) # default to id
        id_key = self.loader_information.get('id_key', 'id') # default to id
        selector = self.selector
        embed_keys = selector.to_embed
        metadata_keys = selector.to_metadata

        with open(file.file_path, newline="", encoding=encoding) as csvfile:
            csv_reader = csv.DictReader(csvfile, **csv_args)  # Use csv_args if provided
            for i, row in enumerate(csv_reader):
                document_id = f"{row.get(id_key, '')}.{id_key}"
                metadata = self.extract_metadata(row)
                content = self.extract_content(row)
                source = row[source_column] if source_column else file.file_path
                metadata["source"] = source
                metadata["row"] = i
                doc = NeumDocument(content=content, metadata=metadata, id=document_id)
                yield doc

    def extract_metadata(self, row: dict) -> dict:
        # Adapted from CSVLoader.extract_metadata()
        return {key: row[key] for key in self.selector.to_metadata if key in row}
    
    def extract_content(self, row: dict) -> str:
        # Content extraction logic goes here
        if self.selector.to_embed:
            row = {k: row[k] for k in self.selector.to_embed if k in row}
        return "\n".join(f"{k.strip()}: {v.strip()}" for k, v in row.items())
